# -*- coding: utf-8 -*-

import sys
import os
import re
import urllib
import urllib2
import cookielib
import MultipartPostHandler as mpph
import time
import zipfile
import getpass
import random
import traceback

import config
from version import VERSION

def main():
    try:
        Main()
    except Exception:
        i=1
        fname="traceback.1.txt"
        while os.path.exists(fname):
            i+=1
            fname="traceback."+str(i)+".txt"
        outp=file(fname,"w")
        traceback.print_exc(file=outp)
        outp.close()
        raw_input("Something unanticipated happened, traceback logged to file '"+fname+"'.")

class Main:
    def __init__(self):
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        self.version=".".join([str(n) for n in VERSION])
        self.games=[]
        cj=cookielib.CookieJar()
        self.urlopener=urllib2.build_opener(urllib2.HTTPCookieProcessor(cj),
                                mpph.MultipartPostHandler)
        self.interface=config.INTERFACE_CLASS(self)
        if not self.config_exists():
            self.interface.on_firststart()
            while True:
                self.prompt_gen_config()
                res=self.login()
                if not res:
                    self.interface.on_badlogin()
                    if self.interface.prompt_quit():
                        sys.exit()
                    continue
                break
            #self.select_tournament()
        else:
            self.load_config()
            res=self.login()
            if not res:
                self.interface.on_logindata_outdated()
                while True:
                    res=self.interface.prompt_reconfig()
                    if res=="recf":
                        self.prompt_gen_config()
                        #and jump back to self.prompt_gen_config(), etc...
                    if res=="newp":
                        self.prompt_gen_password()
                    else:
                        sys.exit()
                    res=self.login()
                    if res:
                        self.interface.on_login_successful()
                        break
                    self.interface.on_connected_login_fail()
            else:
                self.interface.on_stored_login_success()
        self.select_tournament()
        self.get_details()
        self.wait_get_key()
        self.setup_vsweep()
        self.launch_vsweep()
        self.mainloop()

    def select_tournament(self):
        (self.tournament_id,self.tournament_name)=self.prompt_tournament(
                                                    self.get_tournament_list())

    def wait_get_key(self):
        self.interface.on_get_key()
        while True:
            request=urllib2.Request(
                        config.SERVER_BASE_URL+"clientapi/get_key/"
                                            +str(self.tournament_id))
            response=self.urlopener.open(request)
            try:
                message=response.read()
                (code,reason,extradata)=self.parse_message(message)
                if not code and reason=="no_key":
                    time.sleep(3)
                    continue
                break
            except urllib2.URLError:
                pass #handled below
            except ScoreganizerAPIException, ValueError:
                #a likely cause, the user is most likely not talking to the real
                # server here.
                self.connection_warning()
            self.cannot_connect()
        self.tournament_key=extradata
        self.interface.on_got_key()

    def get_details(self):
        #later, get details, especially the starting time. But for now, just
        # try to get a key immediately.
        #print "Getting details for selected tournament..."
        return
        while True:
            request=urllib2.Request(
                        config.SERVER_BASE_URL+"clientapi/detail/"
                                            +str(self.tournament_id)+"/")
            response=self.urlopener.open(request)
            try:
                message=response.read()
                (code,reason,extradata)=self.parse_message(message)
                try:
                    pass
                except ValueError:
                    raise ScoreganizerAPIException("Malformed response")
            except urllib2.URLError:
                pass #handled below
            except ScoreganizerAPIException, ValueError:
                #a likely cause, the user is most likely not talking to the real
                # server here.
                self.connection_warning()
            self.cannot_connect()

    def prompt_tournament(self,tournament_list,more=False):
        res=self.interface.prompt_tournament(tournament_list,more=False)
        if res=="m":
            return self.prompt_tournament(self.get_tournament_list(
                                                        all_active=True),
                                                        more=True)
        if not res:
            sys.exit()
        return res

    def get_tournament_list(self,all_active=False):
        self.interface.on_get_tournament_list()
        while True:
            prefix="all_" if all_active else ""
            request=urllib2.Request(
                        config.SERVER_BASE_URL+"clientapi/"+prefix+"active/")
            response=self.urlopener.open(request)
            try:
                message=response.read()
                (code,reason,extradata)=self.parse_message(message)
                extradata=extradata or ""
                try:
                    tournaments=[(int(t_id),name) for (t_id,name) in
                                    [tuple(line.split(":",1)) for line in
                                        [line for line in
                                            [line.strip() for line in extradata.split("\n")]
                                        if line] # and ":" in line]
                                    ]
                                ]
                except ValueError:
                    raise ScoreganizerAPIException("Malformed response")
                return tournaments
            except urllib2.URLError:
                pass #handled below
            except ScoreganizerAPIException, ValueError:
                #a likely cause, the user is most likely not talking to the real
                # server here.
                self.connection_warning()
            self.cannot_connect()

    def mainloop(self):
        self.interface.on_mainloop()
        while True:
            #checking for new scores every second is more than enough
            time.sleep(1)
            self.check_upload()
        self.interface.on_mainloop_end()

    def login(self):
        while True:
            request=urllib2.Request(config.SERVER_BASE_URL+"clientapi/login/",
                                        urllib.urlencode(
                                                {
                                                    "username":self.username,
                                                    "password":self.password,
                                                    "version":self.version
                                                }
                                            )
                                        )
            try:
                response=self.urlopener.open(request)
                message=response.read()
                (code,reason,extradata)=self.parse_message(message)
                res=bool(code)
            except urllib2.URLError as e:
                res=None
            except ScoreganizerAPIException:
                self.connection_warning()
                res=None
            if res==0 and reason=="need_https":
                self.interface.on_need_https()
                sys.exit(0)
            if res==0 and reason=="client_incompatible":
                self.interface.on_client_incompatible()
                sys.exit(0)
            if res is None:
                self.cannot_connect()
                continue
            return res

    def connection_warning(self):
        #a likely cause, the user is most likely not talking to the real
        # server here.
        self.interface.connection_warning()

    def cannot_connect(self):
        self.interface.on_cannot_connect()
        self.retry_or_quit()

    def retry_or_quit(self):
        res=self.interface.prompt_retry_or_quit()
        if not res:
            sys.exit()

    def parse_message(self,message):
        res=re.split("[ \\n]",message,2)
        if len(res)==1:
            raise ScoreganizerAPIException("Malformed response")
        try:
            res[0]=int(res[0])
        except ValueError:
            raise ScoreganizerAPIException("Malformed response")
        if len(res)==2:
            return tuple(res)+(None,)
        return res

    def check_upload(self):
        games=self.get_games()
        if not games:
            return
        for game in reversed(games):
            if game in self.games:
                break
            self.upload(game)
        self.games=games

    def get_minesweeper_dir(self):
        return "minesweeper-"+self.tournament_key+os.sep

    def upload(self,game):
        vidfname=self.get_minesweeper_dir()+"vids"+os.sep+"archive"\
                                                    +os.sep+game["vidfname"]
        self.interface.on_upload(vidfname)
        ioerrs=0
        while True:
            request=urllib2.Request(config.SERVER_BASE_URL+"clientapi/upload/",
                                            urllib.urlencode(
                                                    {
                                                    }
                                                )
                                            )
            try:
                response=self.urlopener.open(request,{"video":open(vidfname,"rb")})
                message=response.read()
                (code,reason,extradata)=self.parse_message(message)
                res=bool(code)
            except urllib2.URLError as e:
                res=None
            except ScoreganizerAPIException:
                self.connection_warning()
                res=None
            except IOError as e:
                if ioerrs<=4:
                   ioerrs+=1
                   time.sleep(random.randrange(1,10)/16.0)
                   continue
                raise e
            if res is None:
                self.cannot_connect()
                #continue
            elif not res and reason=="retry":
                time.sleep(random.randrange(4)/10.0)
                self.interface.on_retrying()
                continue
            elif not res and reason=="tournament_over":
                self.interface.on_tournament_over()
            elif not res:
                self.interface.on_invalid_data()
            break
        return res

    def get_games(self):
        fname="minesweeper-"+self.tournament_key+os.sep+"vids"+os.sep+self.username+".csv"
        if not os.path.exists(fname):
            return []
        e=None
        for try_ in xrange(5):
            try:
                inp=file(fname,"r")
                lines=[line.strip() for line in inp.readlines() if line.strip()]
                inp.close()
            except IOError as e:
                time.sleep(random.randrange(1,10)/16.0)
                continue
            break
        if not e is None:
            raise e
        colnames=("name","nick","archive","vidfname","bbbv","level","nf","hundredths","timestamp")
        games=[]
        for line in lines:
            cols=[col for col in line.split(",") if col]
            linedict={}
            for col,val in zip(colnames,cols):
                linedict[col]=val
            if linedict["nf"]=="NF": #new history format
                                    #will never happen with regular lines
                                    #as this column is always "0" or "1"
                continue
            games.append(linedict)
        return games

    def prompt_gen_config(self):
        self.username=self.interface.prompt_username()
        self.prompt_gen_password()
        #TODO: Possibly give the user the opportunity to configure their
        # country. This is less important as the country doesn't show up
        # in videos anyway.
        #The implementation below is not complete.
        #while True:
            #print "What country are you playing for?\n"
            #inp=""
            #while True:
                #inp+=getch.getch()
                #matches=[]
                #for country in countries.COUNTRIES_PLUS:
                    #match=re.search(inp.lower(),country[1].lower())
                    #if match:
                        #matches.append(country)
                #if not matches:
                    #print ""
            #country=raw_input("What country are you playing for?\n")
            #if not self.validate_country(country):
                #not pretty, but will work for the moment
                #also, I don't know vsweeps requirements, and ASCII is bound to
                #work
                #print "Country contained invalid characters, please use ASCII."
                #continue
            #break

    def prompt_gen_password(self):
        self.password=self.interface.prompt_password()
        self.save_config(self.username,self.password) #for now
        #this is invoked either as a part of prompt_gen_config or standalone,
        # but either way we want to save the new config.

    def validate_username(self,username):
        return not re.match("[\\w.@+-]{1,30}",username) is None

    def save_config(self,username,password):
        outp=file(".scoreganizer-userdata","w")
        outp.write(username+":"+password+"\n")
        outp.close()

    def load_config(self):
        inp=file(".scoreganizer-userdata","r")
        (self.username,self.password)=tuple(inp.read().strip().split(":",1))

    def config_exists(self):
        return os.path.exists(".scoreganizer-userdata")

    def setup_vsweep(self):
        if os.path.exists("minesweeper-"+self.tournament_key+os.sep):
            #vsweep already set up
            return
        tplinifile=open("initemplate.ini","r")
        inicontent=tplinifile.read()
        tplinifile.close()
        zf=zipfile.ZipFile("vsweep.zip")
        zf.extractall(path=config.EXTRACTPATH)
        zf.close()
        inicontent=inicontent.replace("$$$NAME$$$",self.username)\
                            .replace("$$$NICKNAME$$$",self.tournament_key)\
                            .replace("$$$COUNTRY$$$","Sweeperland")
        vinifile=open("minesweeper"+os.sep+"vsweep"+os.sep+"vienna.ini","w")
        vinifile.write(inicontent)
        vinifile.close()
        os.rename("minesweeper"+os.sep,"minesweeper-"+self.tournament_key+os.sep)

    def launch_vsweep(self):
        msd=self.get_minesweeper_dir()
        cwd=os.getcwd()
        os.chdir(msd)
        try:
            if os.name=="nt":
                os.startfile("minesweeper.bat")
                os.chdir(cwd)
                return
            elif os.name=="posix":
                os.chdir("vsweep")
                os.spawnlp(os.P_NOWAIT,"wine","wine","vsweep.exe")
                os.chdir(cwd)
                return
            e=None
        except Exception as e:
            pass #and store in e
        finally:
            os.chdir(cwd)
        self.interface.on_vsweep_spawn_fail(e)

class ScoreganizerAPIException(Exception):
    pass


if __name__=="__main__":
    main()

